#pragma once

// [scm-mem.h]
// -------------------------------------------------------------------------------------------- //
// memory handling, array ops & simple math
// -------------------------------------------------------------------------------------------- //


#include <stdlib.h>
#include <math.h>
#if defined(_OPENMP)
#include <omp.h>
#endif
#include <time.h>
#include "scm-sys.h"


namespace scm
{
	// * memory (de)allocation * // 
	// -------------------------------------------------------------------------------------------- //
	template< typename T >
	bool allocate(T** x, const int n);
	template< typename T >
	bool allocate(T** x, T** y, const int n);
	template< typename T >
	bool allocate(T** x, T** y, T** z, const int n);

	template< typename T >
	void deallocate(T* x);
	template< typename T >
	void deallocate(T* x, T* y);
	template< typename T >
	void deallocate(T* x, T* y, T* z);

	template< typename T >
	bool allocate_preset(T** x, const T value, const int n);
	template< typename T >
	bool allocate_preset(T** x, T** y, const T value, const int n);
	template< typename T >
	bool allocate_preset(T** x, T** y, T** z, const T value, const int n);
	// -------------------------------------------------------------------------------------------- //

	// * swap * //
	template< typename T >
	void swap(T& a, T& b);
	// -------------------------------------------------------------------------------------------- //

	// * math & vector operations * //
	// -------------------------------------------------------------------------------------------- //
	template< typename T >
	T min(const T a, const T b);
	template< typename T >
	T max(const T a, const T b);
	template< typename T >
	T min(const T a, const T b, const T c);
	template< typename T >
	T max(const T a, const T b, const T c);
	template< typename T >
	T min(const T a, const T b, const T c, const T d);
	template< typename T >
	T max(const T a, const T b, const T c, const T d);

	template< typename T >
	T sgn(const T a);

	template< typename T >
	void assign(T* _RESTRICT x, const T value, const int n);
	template< typename T >
	void null(T* _RESTRICT x, const int n);
	template< typename T >
	void null(T* _RESTRICT x, T* _RESTRICT y, const int n);
	template< typename T >
	void null(T* _RESTRICT x, T* _RESTRICT y, T* _RESTRICT z, const int n);

	template< typename T >
	T max(const T* _RESTRICT x, const int n);
	template< typename T >
	T min(const T* _RESTRICT x, const int n);
	template< typename T >
	T avg(const T* _RESTRICT x, const int n);
	template< typename T >
	T intsum(const T* _RESTRICT f, const T* _RESTRICT dx, const int n);

	template< typename T >
	T interp(const T* _RESTRICT X, const T* _RESTRICT z, const int n, const T pz);
	// -------------------------------------------------------------------------------------------- //
	
	// * vector checks * //
	// -------------------------------------------------------------------------------------------- //
	template< typename T >
	bool is_finite(const T* _RESTRICT x, const int n);
	// -------------------------------------------------------------------------------------------- //

	// * timer * //
	// -------------------------------------------------------------------------------------------- //
	double get_sys_time();
	// -------------------------------------------------------------------------------------------- //
}


// -------------------------------------------------------------------------------------------- //
// * Implementation *
// -------------------------------------------------------------------------------------------- //

// * memory (de)allocation * //
// -------------------------------------------------------------------------------------------- //
template< typename T >
inline bool scm::allocate(
	T** x, const int n)
{
#ifndef ALIGN_ALLOCATION
	(*x) = new T[n];
#else
#if defined(__INTEL_COMPILER)
	(*x) = (T*)_mm_malloc(n * sizeof(T), ALIGN_ALLOCATION);
	if (*x == NULL) return false;

#elif defined(_MSC_VER)
	(*x) = (T*)_aligned_malloc(n * sizeof(T), ALIGN_ALLOCATION);
	if (*x == NULL) return false;

#else // die or use posix //
	if (posix_memalign((void**)x, ALIGN_ALLOCATION, n * sizeof(T)) != 0)
		return false;
#endif
#endif

	return true;
}

template< typename T >
inline bool scm::allocate(
	T** x, T** y, const int n)
{
	if (!allocate(x, n)) return false;
	if (!allocate(y, n)) {
		deallocate(*x);
		return false;
	}

	return true;
}

template< typename T >
inline bool scm::allocate(
	T** x, T** y, T** z, const int n)
{
	if (!allocate(x, n)) return false;
	if (!allocate(y, n)) {
		deallocate(*x);
		return false;
	}
	if (!allocate(z, n)) {
		deallocate(*x);
		deallocate(*y);
		return false;
	}

	return true;
}

template< typename T >
inline void scm::deallocate(
	T* x)
{
#ifndef ALIGN_ALLOCATION
	delete[] x;
#else
#if defined(__INTEL_COMPILER)
	_mm_free((void*)x);
#elif defined(_MSC_VER)
	_aligned_free((void*)x);
#else	// die or use posix //
	free((void*)x);
#endif
#endif
}

template< typename T >
inline void scm::deallocate(
	T* x, T* y)
{
	deallocate(x);
	deallocate(y);
}

template< typename T >
inline void scm::deallocate(
	T* x, T* y, T* z)
{
	deallocate(x);
	deallocate(y);
	deallocate(z);
}

template< typename T >
inline bool scm::allocate_preset(
	T** x, const T value, const int n)
{
	if (!allocate(x, n)) return false;

	assign((*x), value, n);
	return true;
}

template< typename T >
inline bool scm::allocate_preset(
	T** x, T** y, const T value, const int n)
{
	if (!allocate(x, y, n)) return false;

	assign((*x), value, n);
	assign((*y), value, n);
	return true;
}

template< typename T >
inline bool scm::allocate_preset(
	T** x, T** y, T** z, const T value, const int n)
{
	if (!allocate(x, y, z, n)) return false;

	assign((*x), value, n);
	assign((*y), value, n);
	assign((*z), value, n);
	return true;
}
// -------------------------------------------------------------------------------------------- //

// * swap * //
// -------------------------------------------------------------------------------------------- //
template< typename T >
inline void scm::swap(T& a, T& b)
{
	T c = a; a = b; b = c;
}
// -------------------------------------------------------------------------------------------- //

// * math & vector operations * //
// -------------------------------------------------------------------------------------------- //
template< typename T >
inline T scm::min(
	const T a, const T b)
{
	return (a < b) ? a : b;
}

template< typename T >
inline T scm::max(
	const T a, const T b)
{
	return (a > b) ? a : b;
}

template< typename T >
inline T scm::min(
	const T a, const T b, const T c)
{
	return (a < b) ? ((a < c) ? a : c) : ((b < c) ? b : c);
}

template< typename T >
inline T scm::max(
	const T a, const T b, const T c)
{
	return (a > b) ? ((a > c) ? a : c) : ((b > c) ? b : c);
}

template< typename T >
inline T scm::min(
	const T a, const T b, const T c, const T d)
{
	return min(min(a, b, c), d);
}

template< typename T >
inline T scm::max(
	const T a, const T b, const T c, const T d)
{
	return max(max(a, b, c), d);
}

template< typename T >
inline T scm::sgn(
	const T a)
{
	if (a < (T)0) return (T)-1.0;
	if (a > (T)0) return (T)1.0;

	return (T)0;
}

template< typename T >
inline void scm::assign(
	T* _RESTRICT x, const T value, const int n)
{
#pragma omp parallel shared( x )
	{
		int i;

#pragma omp for nowait
		for (i = 0; i < n - (n % 4); i += 4) {
			x[i] = value;
			x[i + 1] = value;
			x[i + 2] = value;
			x[i + 3] = value;
		}

#pragma omp single nowait
		for (i = n - (n % 4); i < n; i++)
			x[i] = value;
	}
}

template< typename T >
inline void scm::null(
	T* _RESTRICT x, const int n)
{
#pragma omp parallel shared( x )
	{
		int i;

#pragma omp for nowait
		for (i = 0; i < n - (n % 4); i += 4) {
			x[i] = (T)0;
			x[i + 1] = (T)0;
			x[i + 2] = (T)0;
			x[i + 3] = (T)0;
		}

#pragma omp single nowait
		for (i = n - (n % 4); i < n; i++)
			x[i] = (T)0;
	}
}

template< typename T >
inline void scm::null(
	T* _RESTRICT x, T* _RESTRICT y, const int n)
{
	null(x, n);
	null(y, n);
}

template< typename T >
inline void scm::null(
	T* _RESTRICT x, T* _RESTRICT y, T* _RESTRICT z, const int n)
{
	null(x, n);
	null(y, n);
	null(z, n);
}

template< typename T >
inline T scm::min(
	const T* _RESTRICT const x, const int n)
{
	if (n <= 0) return (T)0;

	int i;
	T _min = (T)(*x);

	T _min_local = (T)(*x);

#pragma omp parallel firstprivate(_min_local) private(i) shared(_min)
	{
#pragma omp for nowait
		for (i = 1; i < n; i++) {
			if (x[i] < _min_local) _min_local = x[i];
		}

#pragma omp critical
		{
			if (_min_local < _min) _min = _min_local;
		}
	}

	return _min;
}

template< typename T >
inline T scm::max(
	const T* _RESTRICT const x, const int n)
{
	if (n <= 0) return (T)0;

	int i;
	T _max = (T)(*x);

	T _max_local = (T)(*x);

#pragma omp parallel firstprivate(_max_local) private(i) shared(_max)
	{
#pragma omp for nowait
		for (i = 1; i < n; i++) {
			if (x[i] > _max_local) _max_local = x[i];
		}

#pragma omp critical
		{
			if (_max_local > _max) _max = _max_local;
		}
	}

	return _max;
}

template< typename T >
inline T scm::avg(
	const T* _RESTRICT const x, const int n)
{
	if (n <= 0) return (T)0;

	int i;
	T _sum = (T)0;

#pragma omp parallel for private( i ) reduction( + : _sum )
	for (i = 0; i < n - (n % 4); i += 4) {
		_sum += x[i] + x[i + 1] + x[i + 2] + x[i + 3];
	}

	for (i = n - (n % 4); i < n; i++)
		_sum += x[i];

	return _sum / n;
}

template< typename T >
inline T scm::intsum(
	const T* _RESTRICT const f, 
	const T* _RESTRICT const dx,
	const int n)
{
	if (n <= 0) return (T)0;

	int i;
	T _sum = (T)0;

#pragma omp parallel for private( i ) reduction( + : _sum )
	for (i = 0; i < n - (n % 4); i += 4) {
		_sum += 
			f[i] * dx[i] + f[i + 1] * dx[i + 1] + 
			f[i + 2] * dx[i + 2] + f[i + 3] * dx[i + 3];
	}

	for (i = n - (n % 4); i < n; i++)
		_sum += f[i] * dx[i];

	return _sum;
}

template< typename T >
inline T scm::interp(
	const T* _RESTRICT X, 
	const T* _RESTRICT z, const int n, 
	const T pz)
{
	if (n == 0) return (T)0;
	if (n == 1) return X[0];

	int k;

	k = 0;
	if (pz < z[k]) {
		// --- extrapolation, [zpos, zpos + dz/2):
		T alpha = (z[k + 1] - pz) / (z[k + 1] - z[k]);
		return alpha * X[k] + ((T)1.0 - alpha) * X[k + 1];
	}

	k = n - 1;
	if (pz > z[k]) {
		// --- extrapolation, (zpos + height - dz/2, zpos + height]
		T alpha = (z[k] - pz) / (z[k] - z[k - 1]);
		return alpha * X[k - 1] + ((T)1.0 - alpha) * X[k];
	}

	for (k = 0; k < n - 1; k++) {
		if ((pz >= z[k]) && (pz <= z[k + 1]))
		{
			T alpha = (z[k + 1] - pz) / (z[k + 1] - z[k]);
			return alpha * X[k] + ((T)1.0 - alpha) * X[k + 1];
		}
	}

	return (T)0;
}
// -------------------------------------------------------------------------------------------- //

// * vector checks * //
// -------------------------------------------------------------------------------------------- //
template< typename T >
inline bool scm::is_finite(
	const T* _RESTRICT x, const int n)
{
	int i;
	int num_inf = 0;

#pragma omp parallel for private( i ) reduction( + : num_inf )
	for (i = 0; i < n - (n % 4); i += 4) {
		if ((!isfinite(x[i])) || (!isfinite(x[i + 1])) ||
			(!isfinite(x[i + 2])) || (!isfinite(x[i + 3]))) num_inf++;
	}

	for (i = n - (n % 4); i < n; i++)
		if (!isfinite(x[i])) num_inf++;

	return (num_inf == 0);
}
// -------------------------------------------------------------------------------------------- //

// * timer * //
// -------------------------------------------------------------------------------------------- //
inline double scm::get_sys_time()
{
#if defined(_OPENMP)
	return omp_get_wtime();
#else
	return ((double)clock() / CLOCKS_PER_SEC);
#endif
}
// -------------------------------------------------------------------------------------------- //

